% Demo code for the paper:
% Soft clustering for nonparametric probability density function estimation
% by Ezequiel Lpez-Rubio and Juan Miguel Ortiz-de-Lazcano-Lobato
% Pattern Recognition Letters. ISSN: 0167-8655. DOI: 10.1016/j.patrec.2008.07.010 
% Coded by Ezequiel Lpez-Rubio, April 2012.

clear all

NumSamples=1000;
NumClusters=900;

disp('Smooth Parzen Windows demo');
MethodNumber=input('Please specify the method to use (1=dist, 2=fuzzy): ');
switch MethodNumber
    case 1
        VarianceAmount=0.9;
        NumNeighbors=10;
        GammaParameter=0.7;
        PhiParameter=0.05; 
        Method='dist';
    case 2
        VarianceAmount=0.1;
        NumNeighbors=10;
        GammaParameter=0.5;
        PhiParameter=nan; % Not used for the 'fuzzy' method
        Method='fuzzy';
    otherwise
        error('Unknown method');
end
        
FileNumber=input('Please specify the input dataset (1=Ball, 2=ThickS, 3=ThinS, 4=ThreeBalls, 5=TwoShapes, 6=X): ');
ListFileNames={'Ball.jpg','ThickS.jpg','ThinS.jpg','ThreeBalls.jpg','TwoShapes.jpg','X.jpg'};
if ismember(FileNumber,1:numel(ListFileNames))
    FileName=ListFileNames{FileNumber};
else
    error('Invalid dataset');
end

disp('Generating training samples...')
Samples=GenerateSamplesImg(FileName,NumSamples);

disp('Building the model...')
[Model]=TrainSmoothParzenWindows(Samples,VarianceAmount,NumNeighbors,NumClusters,PhiParameter,GammaParameter,Method);

% Measure the performance
disp('Generating test samples...')
TestSamples=GenerateSamplesImg(FileName,10000);
disp('Measuring ANLL...')
TestDensity=TestSmoothParzenMEX(Model,TestSamples);
ANLL=-(1/size(TestSamples,2))*sum(log(TestDensity))

disp('Preparing the plots...')
[X Y]=meshgrid(0:0.003:1,0:0.003:1);
TestSamplesPlot=zeros(2,numel(X));
TestSamplesPlot(1,:)=reshape(X,1,numel(X));
TestSamplesPlot(2,:)=reshape(Y,1,numel(X));

DensitiesPlot=TestSmoothParzenMEX(Model,TestSamplesPlot);
fprintf('Estimated integral of the probability density in the plotted area: %f\n',sum(DensitiesPlot(:))/(334*334));


plot(Samples(1,:),Samples(2,:),'xg')
hold on
title('Training samples')
axis square

figure
imagesc(flipud(reshape(DensitiesPlot,334,334)))
hold on
title('Estimated probability density')
colormap(flipud(hot))
colorbar
axis off
