#include "mex.h"
#include <math.h>

/* 

This is part of the demo code for the paper:
	Soft clustering for nonparametric probability density function estimation
		by Ezequiel Lpez-Rubio and Juan Miguel Ortiz-de-Lazcano-Lobato
	Pattern Recognition Letters. ISSN: 0167-8655. DOI: 10.1016/j.patrec.2008.07.010 
	Coded by Ezequiel Lpez-Rubio, April 2012.

In order to compile this MEX function, type the following at Matlab prompt:
>> mex TestSmoothParzenMEX.c


[Densities]=TestSmoothParzenMEX(Model,TestSamples);

*/

/* Find the squared Euclidean norm of a vector */
void SquaredNorm(double * const Vector,double * const Result,int Dimension)
{
    register double *MyComponent;
    register int ndx;
    register double MyResult;    
    
    MyComponent=Vector;
    MyResult=0.0;
    for (ndx=0;ndx<Dimension;ndx++)
    {
        MyResult+=(*MyComponent)*(*MyComponent);
        MyComponent++;                
    }    
    (*Result)=MyResult;  
}

/* Matrix product */
void MatrixProduct(double *A,double *B,double *Result,int NumRowsA,
    int NumColsA,int NumColsB)
{
    register double *ptra;
    register double *ptrb;
    register double *ptrres;
    register int i;
    register int j;
    register int k;
    register double Sum;
    
    ptrres=Result;
    for(j=0;j<NumColsB;j++)
    {
        for(i=0;i<NumRowsA;i++)
        {
            Sum=0.0;
            ptrb=B+NumColsA*j;
            ptra=A+i;
            for(k=0;k<NumColsA;k++)
            {
                Sum+=(*ptra)*(*ptrb);
                ptra+=NumRowsA;
                ptrb++;
            }    
            (*ptrres)=Sum;
            ptrres++;
        }
    }            
}   

/* Matrix difference */
void MatrixDifference(double *A,double *B,double *Result,int NumRows,int NumCols)
{
    register double *ptra;
    register double *ptrb;
    register double *ptrres;
    register int ndx;
    register int NumElements;
    
    ptra=A;
    ptrb=B;
    ptrres=Result;
    NumElements=NumRows*NumCols;
    for(ndx=0;ndx<NumElements;ndx++)
    {
        (*ptrres)=(*ptra)-(*ptrb);
        ptrres++;
        ptra++;
        ptrb++;
    }    
}



void mexFunction(int nlhs, mxArray* plhs[],
                 int nrhs, const mxArray* prhs[])
{   
    double *ptrOutput,*DifferenceVector,*ptrSigma2,*VectorAux;
    double *ptrMyMean,*ptrMyEigenvalues,*ptrMyEigenvectors;
    mxArray *Means,*Eigenvalues,*Eigenvectors,*NumPrincipalDirections,*Sigma2;
    mxArray *NormCovariances,*InvCovariances;
    double *ptrTestSamples,*ptrMyTestSample,*ptrMyNormCovariances,*ptrMyInvCovariances;
    double MySigma2,ConstantTerm,r,q,AuxDensity;
    const int *DimTestSamples;
    double ScalarProduct,VarianceAmount,ConstantFactor;
    int SpaceDimension,NumSamplesTest,NumKernels,MyNumPrincipalDirections;
    int NdxTestSample,NdxKernel,NdxDimension;


    /* Get working variables */
    ptrTestSamples=mxGetPr(prhs[1]);
    Means=mxGetField(prhs[0],0,"Means");
    VarianceAmount=(*mxGetPr(mxGetField(prhs[0],0,"VarianceAmount")));
    NumKernels=(int)(*mxGetPr(mxGetField(prhs[0],0,"NumClusters")));
    DimTestSamples=mxGetDimensions(prhs[1]);
    SpaceDimension=DimTestSamples[0];
    NumSamplesTest=DimTestSamples[1];

    /* See if we have used the explained variance method or full covariance matrices */
    if (VarianceAmount==0.0)
    {
        NormCovariances=mxGetField(prhs[0],0,"NormCovariances");
        InvCovariances=mxGetField(prhs[0],0,"InvCovariances");
        /* Precompute ConstantFactor= (2*pi)^(-Dimension/2) */
        ConstantFactor=pow(6.283185307179586,-0.5*(double)SpaceDimension);
    }
    else
    {
        Eigenvalues=mxGetField(prhs[0],0,"EigenValues");
        Eigenvectors=mxGetField(prhs[0],0,"EigenVectors");
        Sigma2=mxGetField(prhs[0],0,"Sigma2");
        NumPrincipalDirections=mxGetField(prhs[0],0,"NumPrincipalDirections");
    }        
    
    
    
    /* Auxiliary matrices */
    DifferenceVector=mxMalloc(sizeof(double)*SpaceDimension);
    VectorAux=mxMalloc(sizeof(double)*SpaceDimension);
    
    /* Create output matrix */
    plhs[0]=mxCreateDoubleMatrix(NumSamplesTest,1,mxREAL);
    ptrOutput=mxGetPr(plhs[0]);

     

    for(NdxTestSample=0;NdxTestSample<NumSamplesTest;NdxTestSample++)
    {
        /* Density=0; */
        ptrOutput[NdxTestSample]=0.0;
   
        ptrMyTestSample=ptrTestSamples+NdxTestSample*SpaceDimension;
        
                              
        for(NdxKernel=0;NdxKernel<NumKernels;NdxKernel++)
        {
            /* Find the difference vector with respect to the mean */
            ptrMyMean=mxGetPr(mxGetCell(Means,NdxKernel));  
            MatrixDifference(ptrMyTestSample,ptrMyMean,DifferenceVector,SpaceDimension,1);
                        
            /* See if we have used the explained variance method or full covariance matrices */
            if (VarianceAmount==0.0)
            {
				/* Full covariance matrix */

                /* Get pointers to the parameters of this kernel */
                ptrMyNormCovariances=mxGetPr(mxGetCell(NormCovariances,NdxKernel));
                ptrMyInvCovariances=mxGetPr(mxGetCell(InvCovariances,NdxKernel));
                
                /* Density=(2*pi)^(-Dimension/2)*(Model.NormCovariances{NdxGaussian}^(-0.5)); */
                AuxDensity=ConstantFactor*pow(*ptrMyNormCovariances,-0.5);
                
                /* Density=Density*exp(-0.5*VectorDif'*Model.InvCovariances{NdxGaussian}*VectorDif); */
                MatrixProduct(DifferenceVector,ptrMyInvCovariances,VectorAux,1,SpaceDimension,SpaceDimension);
                MatrixProduct(VectorAux,DifferenceVector,&ScalarProduct,1,SpaceDimension,1);
                ptrOutput[NdxTestSample]+=AuxDensity*exp(-0.5*ScalarProduct);
                      
            }
            else
            {    
				/* Explained variance method */

                /* Get pointers to the parameters of this kernel */
                ptrMyEigenvalues=mxGetPr(mxGetCell(Eigenvalues,NdxKernel));
                ptrMyEigenvectors=mxGetPr(mxGetCell(Eigenvectors,NdxKernel));
                MySigma2=(*mxGetPr(mxGetCell(Sigma2,NdxKernel)));   
                MyNumPrincipalDirections=(int)(*mxGetPr(mxGetCell(NumPrincipalDirections,NdxKernel)));
                
                /* ConstantTerm=SpaceDimension*log(2*pi) */
                ConstantTerm=((double)SpaceDimension)*1.837877066409345;
                
                /* r=(SpaceDimension-Model.NumPrincipalDirections{NdxCluster})*log(2*pi)+...
                    sum(log(Model.EigenValues{NdxCluster}+Model.Sigma2{NdxCluster}))+...
                    (Dimension-Model.NumPrincipalDirections{NdxCluster})*log(Model.Sigma2{NdxCluster}); */
                r=ConstantTerm+((double)(SpaceDimension-MyNumPrincipalDirections))*log(MySigma2);
                
                for(NdxDimension=0;NdxDimension<MyNumPrincipalDirections;NdxDimension++)
                {
                    r+=log(ptrMyEigenvalues[NdxDimension]+MySigma2);
                    
                }    
                      
                /* q=norm(TestInput-Model.Means{NdxCluster})^2)/Model.Sigma2{NdxCluster}; */
                SquaredNorm(DifferenceVector,&q,SpaceDimension);                
				q/=MySigma2;
                
                /* for NdxDimension=1:Model.NumPrincipalDirections{NdxCluster} */
                for(NdxDimension=0;NdxDimension<MyNumPrincipalDirections;NdxDimension++)
                {
                    /* q=q+(1/(sigma^2+lambda_i)-1/sigma^2)*...
                            Model.EigenVectors{NdxCluster}(:,NdxDimension)'*...
                            (TestInput-Model.Means{NdxCluster}); */
                    MatrixProduct(ptrMyEigenvectors+NdxDimension*SpaceDimension,DifferenceVector,
                                    &ScalarProduct,1,SpaceDimension,1);
                    ScalarProduct=ScalarProduct*ScalarProduct;
					q+=(1.0/(MySigma2+ptrMyEigenvalues[NdxDimension])-1.0/MySigma2)*ScalarProduct;
                }   
                      
                /* Density=exp(-0.5*(r+q)); */
                ptrOutput[NdxTestSample]+=exp(-0.5*(r+q));
            }    
 
        }    
              
        /* Density=Density/size(Model.Samples,2);*/
        ptrOutput[NdxTestSample]/=(double)NumKernels;
   
    }    
    
    
    
    /* Release memory */
    mxFree(DifferenceVector);
    mxFree(VectorAux);
}    
